#include "editor.h"

#include <stdio.h>
#include <string.h>

#include <string>

using namespace std;

namespace bbs {

    editor::editor(terminal *term) {
        this->term = term;
        this->wrap_column = term->get_width() - 2;

        this->foreground = CYAN;
        this->background = BLACK;
        this->quote_foreground = LGREY;
        this->toolbar_foreground = WHITE;
        this->toolbar_background = BLUE;

        this->window_top = 6;
        this->window_height = term->get_height() - this->window_top;
        this->toolbar_row = 5;

        this->text = "\r";
        this->top_row = 0;
    }

    editor::~editor() {

    }

    int editor::run() {
        int result = 0, state = 0;

        initalize();

        for (;;) {
            int c = term->get_char();
            if (state == 0) {
                if (c == CTRLZ) {
                    result = 1;
                    break;
                }
                if (c == CTRLK) {
                    int tcol = term->get_column();
                    int trow = term->get_row();

                    term->set_foreground(toolbar_foreground);
                    term->set_background(toolbar_background);
                    term->gotoxy(1, toolbar_row);
                    term->put_string(" ^K");

                    term->gotoxy(tcol, trow);

                    state = 1;
                    continue;
                }
                if (c == CTRLE) {
                    c = UP;
                }
                else if (c == CTRLX) {
                    c = DOWN;
                }
                else if (c == CTRLS) {
                    c = LEFT;
                }
                else if (c == CTRLD) {
                    c = RIGHT;
                }
            }
            else if (state == 1) {
                if (c == CTRLE) {
                    c = PGUP;
                }
                else if (c == CTRLX) {
                    c = PGDN;
                }
                else if (c == CTRLS) {
                    c = HOME;
                }
                else if (c == CTRLD) {
                    c = END;
                }

                int tcol = term->get_column();
                int trow = term->get_row();

                term->set_foreground(toolbar_foreground);
                term->set_background(toolbar_background);
                term->gotoxy(1, toolbar_row);
                term->put_string("\xC4\xC4\xC4");

                term->gotoxy(tcol, trow);
                term->set_foreground(foreground);
                term->set_background(background);

                if (c == CTRLQ || toupper(c) == 'Q') {
                    result = 0;
                    break;
                }

                state = 0;
            }

            try {
                process_char(c);
            } catch(exception & e) {
                row = col = 0;
                index = 0;
                row_index.clear();
                row_size.clear();

                term->set_foreground(foreground);
                term->set_background(background);
                redraw_screen(row, true);

                gotoxy(col, row);
            }
        }

        return result;
    }

    void editor::initalize() {
        const char * TOOLBAR_HELP = " ^X=Down ^E=Up ^S=Left ^D=Right ^Z=Save ";

        term->set_foreground(toolbar_foreground);
        term->set_background(toolbar_background);
        term->gotoxy(0, toolbar_row);
        for (int i = 0; i < term->get_width() - strlen(TOOLBAR_HELP) - 1; i++) {
            term->put_char(0xC4);
        }
        term->put_string(TOOLBAR_HELP);
        term->put_char(0xC4);

        row = col = 0;
        index = 0;
        term->set_foreground(foreground);
        term->set_background(background);

        row_index.clear();
        row_size.clear();
        redraw_screen(row, true);

        index = text.length() - 1;
        row = get_row_at(index);
        col = get_col_at(index);

        gotoxy(col, row);
    }

    void editor::process_char(int c) {
        char temp[100];

        switch (c) {
            case UP:
                if (row > 0) {
                    index = get_index(row - 1, col);
                    col = get_col_at(index);
                    row = get_row_at(index);
                    gotoxy(col, row);
                }
                break;
            case DOWN:
                if (row_size[row + 1] > 0) {
                    index = get_index(row + 1, col);
                    col = get_col_at(index);
                    row = get_row_at(index);
                    gotoxy(col, row);
                }
                break;
            case LEFT:
                if (index > 0) {
                    index--;
                    col = get_col_at(index);
                    row = get_row_at(index);
                    gotoxy(col, row);
                }
                break;
            case RIGHT: {
                int margin = text.length();
                if (text[margin - 1] == '\r') {
                    margin--;
                }
                if (index < margin) {
                    index++;
                    col = get_col_at(index);
                    row = get_row_at(index);
                    gotoxy(col, row);
                }
                break;
            }
            case END:
                index += row_size[row] - col - 1;
                col = row_size[row] - 1;
                gotoxy(col, row);
                break;
            case HOME:
                index -= col;
                col = 0;
                gotoxy(col, row);
                break;
            case PGUP:
                row -= window_height - 1;
                if (row < 0) {
                    row = 0;
                }
                top_row -= window_height - 1;
                if (top_row < 0) {
                    top_row = 0;
                }
                index = get_index(row, col);
                row = get_row_at(index);
                col = get_col_at(index);
                redraw_screen(top_row, true);
                gotoxy(col, row);
                break;
            case PGDN: {
                int old_top_row = top_row;
                int old_row = row;

                for (int i = 0; i < window_height - 1; i++) {
                    if (row_size[row + 1] == 0) {
                        top_row = old_top_row;
                        break;
                    }
                    top_row++;
                    row++;
                }

                index = get_index(row, col);
                row = get_row_at(index);
                redraw_screen(top_row, true);
                gotoxy(col, row);
                break;
            }
            case CTRLH: // Backspace
                if (index > 0) {
                    text = text.substr(0, index - 1) + text.substr(index, text.length() - index);
                    index--;

                    if (col > 0) {
                        redraw_screen(row, false);
                        col = get_col_at(index);
                        row = get_row_at(index);
                        gotoxy(col, row);
                    }
                    else if (row > 0) {
                        row--;
                        redraw_screen(row, false);

                        col = get_col_at(index);
                        row = get_row_at(index);
                    }

                    gotoxy(col, row);
                }
                break;
            case DEL: {
                int margin = text.length();
                if (text[margin - 1] == '\r') {
                    margin--;
                }
                if (index < margin) {
                    text = text.substr(0, index) + text.substr(index + 1, text.length() - index - 1);

                    redraw_screen(row, false);
                    col = get_col_at(index);
                    row = get_row_at(index);

                    gotoxy(col, row);
                }
                break;
            }
            case RETURN:
                text.insert(index++, 1, c);
                row_size[row]++;
                redraw_screen(row, false);
                row++;
                col = 0;
                gotoxy(col, row);
                break;
            case CTRLY: {
                int s = index - col;
                text = text.substr(0, s) + text.substr(s + row_size[row], text.length() - row_size[row]);
                redraw_screen(row, false);
                col = 0;
                gotoxy(col, row);
                break;
            }
            default:
                if (c < 32 || c >= 127) {
                    break;
                }
                text.insert(index++, 1, c);
                col++;
                row_size[row]++;

                if (row_size[row] > wrap_column) {
                    redraw_screen(row, false);
                    col = get_col_at(index);
                    row = get_row_at(index);
                    gotoxy(col, row);
                }
                else if (col < row_size[row] && index < text.length() && text[index] != '\r') {
                    const char *p = text.c_str();
                    int len = row_size[row] - col;
                    memcpy(temp, &p[index - 1], len);
                    temp[len] = '\0';
                    term->clreol();
                    term->put_string(temp);
                    gotoxy(col, row);
                }
                else {
                    term->put_char(c);
                }
                break;
        }
    }

    int editor::get_row_at(int index) {
        int r = 0, i = 0;

        for (;;) {
            if ((index - i) < row_size[r] || row_size[r] == 0) {
                break;
            }
            i += row_size[r++];
        }

        return r;
    }

    int editor::get_col_at(int index) {
        int r = 0, i = 0;

        for (;;) {
            if ((index - i) < row_size[r] || row_size[r] == 0) {
                break;
            }
            i += row_size[r++];
        }

        return index - i;
    }

    int editor::get_index(int row, int col) {
        int result = 0;

        for (int i = 0; i < row; i++) {
            result += row_size[i];
        }

        if (col != 0) {
            if (col < row_size[row]) {
                result += col - 1;
            }
            else {
                result += row_size[row] - 1;
            }
            if (text[result] != '\r') {
                result++;
            }
        }

        return result;
    }

    void editor::redraw_screen(int from_row, bool force) {
        int row, index, start_index, height, qi;
        char temp[20];
        string s;

        row = from_row;
        index = row_index[row];

        start_index = index;
        height = from_row - top_row;

        while (index < text.length()) {
            char c = text[index++];
            s += c;

            if (c == '\r') {
                if (height < window_height && (row_size[row] != s.length() || row_index[row] != start_index || force)) {
                    gotoxy(0, row);
                    qi = s.find(">");
                    term->set_foreground((qi != string::npos && qi < 5) ? quote_foreground : foreground);
                    term->put_string(s, s.length() - 1);
                    term->clreol();
                }

                row_size[row] = s.length();
                row_index[row] = start_index;

                row++;
                height++;
                start_index = index;
                s = "";
            }
            else if (s.length() >= wrap_column) {
                int len = s.length() - 1;
                while (len > 0 && s[len] != ' ') {
                    len--;
                    index--;
                }
                if (len > 0) {
                    s = s.substr(0, len + 1);
                }

                if (height < window_height && (row_size[row] != s.length() || row_index[row] != start_index || force)) {
                    gotoxy(0, row);
                    qi = s.find(">");
                    term->set_foreground((qi != string::npos && qi < 5) ? quote_foreground : foreground);
                    term->put_string(s);
                    term->clreol();
                }

                row_size[row] = s.length();
                row_index[row] = start_index;

                row++;
                height++;
                start_index = index;
                s = "";
            }
        }

        if (s.length() > 0) {
            if (height < window_height && (row_size[row] != s.length() || row_index[row] != start_index || force)) {
                gotoxy(0, row);
                qi = s.find(">");
                term->set_foreground((qi != string::npos && qi < 5) ? quote_foreground : foreground);
                term->put_string(s);
                term->clreol();
            }

            row_size[row] = s.length();
            row_index[row++] = start_index;
            height++;
        }

        start_index = index;

        term->set_foreground(foreground);

        while (row_size[row] > 0) {
            if (height < window_height) {
                gotoxy(0, row);
                term->clreol();
            }
            row_size[row] = 0;
            row_index[row++] = start_index;
            height++;
        }

        while (height < window_height) {
            gotoxy(0, row++);
            term->clreol();
            term->put_string(s);
            height++;
        }
    }

    void editor::gotoxy(int x, int y) {
        int old_top_row = top_row;

        if (y < top_row) {
            top_row = y;
        }
        while ((y - top_row) >= window_height) {
            top_row++;
        }

        if (top_row != old_top_row) {
            redraw_screen(top_row, true);
        }

        term->gotoxy(x, window_top + y - top_row);
    }

    void editor::go_to_index(int index) {
        this->index = index;
        this->col = get_col_at(index);
        this->row = get_row_at(index);
        gotoxy(this->col, this->row);
    }

} /* namespace bbs */
